@@ -25,10 +25,15 @@
  * SUCH DAMAGE.
  */
 
+#ifdef WIN32
+#include <windows.h>
+#include <shlwapi.h>
+#endif
+
 #include "file.h"
 
 #ifndef	lint
-FILE_RCSID("@(#)$File: magic.c,v 1.65 2009/09/14 17:50:38 christos Exp $")
+FILE_RCSID("@(#)$File: magic.c,v 1.73 2011/05/10 17:08:14 christos Exp $")
 #endif	/* lint */
 
 #include "magic.h"
@@ -57,8 +62,6 @@ FILE_RCSID("@(#)$File: magic.c,v 1.65 2009/09/14 17:50:38 christos Exp $")
 #include <unistd.h>	/* for read() */
 #endif
 
-#include "patchlevel.h"
-
 #ifndef PIPE_BUF
 /* Get the PIPE_BUF from pathconf */
 #ifdef _PC_PIPE_BUF
@@ -81,26 +84,129 @@ private const char *file_or_fd(struct magic_set *, const char *, int);
 #define	STDIN_FILENO	0
 #endif
 
+#ifdef WIN32
+BOOL WINAPI DllMain(HINSTANCE hinstDLL,
+    DWORD fdwReason __attribute__((__unused__)),
+    LPVOID lpvReserved __attribute__((__unused__)));
+
+CHAR dllpath[MAX_PATH + 1] = { 0 };
+
+BOOL WINAPI DllMain(HINSTANCE hinstDLL,
+    DWORD fdwReason __attribute__((__unused__)),
+    LPVOID lpvReserved __attribute__((__unused__)))
+{
+	if (dllpath[0] == 0 &&
+	    GetModuleFileNameA(hinstDLL, dllpath, MAX_PATH) != 0)
+		PathRemoveFileSpecA(dllpath);
+	return TRUE;
+}
+#endif
+
 private const char *
 get_default_magic(void)
 {
-	static const char hmagic[] = "/.magic";
-	static char default_magic[2 * MAXPATHLEN + 2];
-	char *home;
-	char hmagicpath[MAXPATHLEN + 1];
+	static const char hmagic[] = "/.magic/magic.mgc";
+	static char *default_magic;
+	char *home, *hmagicpath;
 
+#ifndef WIN32
+	struct stat st;
+
+	if (default_magic) {
+		free(default_magic);
+		default_magic = NULL;
+	}
 	if ((home = getenv("HOME")) == NULL)
 		return MAGIC;
 
-	(void)snprintf(hmagicpath, sizeof(hmagicpath), "%s%s", home, hmagic);
-
-	if (access(hmagicpath, R_OK) == -1)
+	if (asprintf(&hmagicpath, "%s/.magic", home) < 0)
 		return MAGIC;
+	if (stat(hmagicpath, &st) == -1)
+		goto out;
+	if (S_ISDIR(st.st_mode)) {
+		free(hmagicpath);
+		if (asprintf(&hmagicpath, "%s/%s", home, hmagic) < 0)
+			return MAGIC;
+		if (access(hmagicpath, R_OK) == -1)
+			goto out;
+	}
 
-	(void)snprintf(default_magic, sizeof(default_magic), "%s:%s",
-	    hmagicpath, MAGIC);
-
+	if (asprintf(&default_magic, "%s:%s", hmagicpath, MAGIC) < 0)
+		goto out;
+	free(hmagicpath);
 	return default_magic;
+out:
+	default_magic = NULL;
+	free(hmagicpath);
+	return MAGIC;
+#else
+	char *hmagicp = hmagicpath;
+	char *tmppath = NULL;
+
+#define APPENDPATH() \
+	do { \
+		if (tmppath && access(tmppath, R_OK) != -1) { \
+			if (hmagicpath == NULL) { \
+				hmagicpath = tmppath; \
+				tmppath = NULL; \
+			} else { \
+				free(tmppath); \
+				if (asprintf(&hmagicp, "%s%c%s", hmagicpath, \
+				    PATHSEP, tmppath) >= 0) { \
+					free(hmagicpath); \
+					hmagicpath = hmagicp; \
+				} \
+			} \
+	} while (/*CONSTCOND*/0)
+				
+	if (default_magic) {
+		free(default_magic);
+		default_magic = NULL;
+	}
+
+	/* First, try to get user-specific magic file */
+	if ((home = getenv("LOCALAPPDATA")) == NULL) {
+		if ((home = getenv("USERPROFILE")) != NULL)
+			if (asprintf(&tmppath,
+			    "%s/Local Settings/Application Data%s", home,
+			    hmagic) < 0)
+				tmppath = NULL;
+	} else {
+		if (asprintf(&tmppath, "%s%s", home, hmagic) < 0)
+			tmppath = NULL;
+	}
+
+	APPENDPATH();
+
+	/* Second, try to get a magic file from Common Files */
+	if ((home = getenv("COMMONPROGRAMFILES")) != NULL) {
+		if (asprintf(&tmppath, "%s%s", home, hmagic) >= 0)
+			APPENDPATH();
+	}
+
+
+	/* Third, try to get magic file relative to dll location */
+	if (dllpath[0] != 0) {
+		if (strlen(dllpath) > 3 &&
+		    stricmp(&dllpath[strlen(dllpath) - 3], "bin") == 0) {
+			if (asprintf(&tmppath,
+			    "%s/../share/misc/magic.mgc", dllpath) >= 0)
+				APPENDPATH();
+		} else {
+			if (asprintf(&tmppath,
+			    "%s/share/misc/magic.mgc", dllpath) >= 0)
+				APPENDPATH();
+			else if (asprintf(&tmppath,
+			    "%s/magic.mgc", dllpath) >= 0)
+				APPENDPATH();
+		}
+	}
+
+	/* Don't put MAGIC constant - it likely points to a file within MSys
+	tree */
+	default_magic = hmagicpath;
+	return default_magic;
+#endif
 }
 
 public const char *
@@ -225,6 +331,14 @@ magic_check(struct magic_set *ms, const char *magicfile)
 	return ml ? 0 : -1;
 }
 
+public int
+magic_list(struct magic_set *ms, const char *magicfile)
+{
+	struct mlist *ml = file_apprentice(ms, magicfile, FILE_LIST);
+	free_mlist(ml);
+	return ml ? 0 : -1;
+}
+
 private void
 close_and_restore(const struct magic_set *ms, const char *name, int fd,
     const struct stat *sb)
@@ -315,7 +429,9 @@ file_or_fd(struct magic_set *ms, const char *inname, int fd)
 		int flags = O_RDONLY|O_BINARY;
 
 		if (stat(inname, &sb) == 0 && S_ISFIFO(sb.st_mode)) {
+#ifdef O_NONBLOCK
 			flags |= O_NONBLOCK;
+#endif
 			ispipe = 1;
 		}
 
